#!/usr/bin/python

import sys, math, traceback, random, copy, pickle, time
import traceback, getopt
import subprocess
import itertools

from common import *

COL_WIDTH = 8 # sirina stolpca pri izpisu

class StanjeIgre:
    __slots__ = ['igralci', 'miza', 'glavni_idx', 'soigralec_idx', 'talon', 'preostanek_talona', 'tip_igre', 'zacetni_idx']
    # igralec =  (socket, ime, roka, pobrane, igra)
    # preostanek_talona je zacasn array za karte iz talona ki ne pripadajo glavnemu igralcu
    # miza je list tuplov, miza[0] je karta k jo je vrgu prvi igralc v TEJ igri, miza[1] pa indeks v stanje.igralci
    # talon je list kart
    # preostanek_talona je set kart ki jih glavni igralec ni pobral
    # tip_igre je izbran tip igre
    # zacetni_idx je indeks zacetnega igralca v trenutni rundi

def najvisja(miza, barva):
	potencialne = [karta for karta in miza if karta[0].barva == barva]
	if len(potencialne) == 0: return None
	return max(potencialne, key = lambda x: x[0].vrednost)[1]

def kdo_je_zmagal(miza):
	ip = najvisja(miza, miza[0][0].barva)
	it = najvisja(miza, TAROK)
	return it if it!=None else ip

def run_game(clients):
	stanje = StanjeIgre()
	
	#random.shuffle(clients)
	#clients.append(clients.pop(0))
	stanje.igralci = [[s,n,[],set([]),TipIgre(-1,False,-1),q] for s,n,q in clients]		# igralci so random zmesani clienti + njihovo stanje

	# izpisemo imena igralcev
	print ' '.join(c[1][:COL_WIDTH].center(COL_WIDTH) for c in stanje.igralci)
	if verbose: print '-'*((COL_WIDTH+1)*4)
	
	# nov deck
	karte = vseKarte()					# random ima periodo 2^6000, 54! =~ 10^71, tko da je kul
	random.shuffle(karte)
	
	# razbitje kart
	stanje.talon = karte[:6]							# talon
	for i in [0,1,2,3]: stanje.igralci[i][2] = set(karte[6+i*12:6+(i+1)*12])	# roke
	
	# kaj bomo igral?
	for i in range(len(stanje.igralci)):
		s,n,r,p,g,cseq = stanje.igralci[i]					# od vseh clientov zberemo predloge za igro
		s.zacni(r)
		g.parse(s.readexpect('PREDLAGAM')[0])
	
	# kdo je zmagu?
	predlogi = [(stanje.igralci[idx][4],4-idx) for idx in range(4)]	# pazi: predlogi[x][2] je _4-idx_, da se bo pravilno sortiral
	predlogi.sort(reverse=True)						# prva je najvec vredna igra, z igralcem na _najnizji_ poziciji
	stanje.glavni_idx = 4-predlogi[0][1]
	stanje.tip_igre = predlogi[0][0]
	glavni = stanje.igralci[stanje.glavni_idx]
	glavni[0].stat['glavni'] += 1

	if verbose:
		print ' '.join((str(stanje.igralci[i][4])+('*' if i==stanje.glavni_idx else '')).center(COL_WIDTH) for i in range(4))
	
	# razbijemo talon
	velikost_grupe = glavni[4].stZalozenihKart
	N_grup = 6/velikost_grupe
	razbitje_talona = [set(stanje.talon[i*velikost_grupe:(i+1)*velikost_grupe]) for i in range(N_grup)]
	
	# kaj bos zalozu?
	t_b = time.time()
	glavni[0].zalozi(list(razbitje_talona))
	ret = glavni[0].readexpect('LEZIM')
	glavni[0].stat['time'] += time.time() - t_b
	pobrane,ret = stringiVKarteN(ret); pobrane = set(pobrane)
	zalozene,ret = stringiVKarteN(ret); zalozene = set(zalozene)
	glavni[2] |= pobrane

	if verbose:
		print "Talon:", '  '.join("(%s)" % " ".join(str(k) for k in grp) + ("*" if grp==pobrane else "") for grp in razbitje_talona)
		print "-"*((COL_WIDTH+1)*4)
	
	if pobrane not in razbitje_talona: raise Exception('Igralec ' + glavni[1] + ' ne zna pobrat talona')
	if not zalozene < glavni[2]: raise Exception('Igralec ' + glavni[1] + ' zalaga karte, ki jih nima')
	if len(pobrane) != velikost_grupe or len(zalozene) != velikost_grupe: raise Exception('Igralec ' + glavni[1] + ' ne zna zalagat (igra %d, pobral %d kart, zalozil %d)' % (velikost_grupe, len(pobrane), len(zalozene)))
	if len(zalozene & set([Karta('%s14' % s) for s in ['K', 'R', 'P', 'S']])) > 0: raise Exception('Igralec ' + glavni[1]  + ' poskusa zalozit kralja.')
	
	glavni[2] -= zalozene
	glavni[3] |= zalozene
	stanje.preostanek_talona = set(stanje.talon).difference(pobrane)
	
	# kdo je soigralec?
	kralj = Karta(stanje.tip_igre.klicaniKralj, 14)
	if stanje.tip_igre.solo:
		stanje.soigralec_idx = stanje.glavni_idx
	else:
		try:
			stanje.soigralec_idx = [(kralj in roka) for roka in [c[2] for c in stanje.igralci]].index(True)
		except:
			stanje.soigralec_idx = stanje.glavni_idx

	# posljemo start
	for i,(s,n,r,p,g,cseq) in enumerate(stanje.igralci):
		t_b = time.time()
		s.igraj(i, stanje.glavni_idx, stanje.tip_igre, stanje.preostanek_talona, pobrane)
		s.readexpect('MHM')
		s.stat['time'] += time.time() - t_b

	timer = time.time()
	stanje.zacetni_idx = 0
	for runda in range(12):
		stanje.miza = []
		
		for iidx in [(i+stanje.zacetni_idx) % 4 for i in range(4)]:		# iidx = indeks v stanje.igralci
			igralec = stanje.igralci[iidx]
			t_b = time.time()
			miza = [x[0] for x in stanje.miza]
			igralec[0].vrzi(stanje.zacetni_idx, miza)
			karta = Karta(igralec[0].readexpect('VRZEM')[0])
			igralec[0].stat['time'] += time.time() - t_b
			if karta not in igralec[2]: raise Exception('Igralec ' + igralec[1] + ' mece karto, ki je nima')
			if karta not in veljavnePoteze(igralec[2], miza): raise Exception('Igralec ' + igralec[1] + ' igra z neveljavno karto: ' + str(karta) + ' na ' + str(miza) + ' s kartami ' + str(igralec[2]))
			igralec[2].remove(karta)
			stanje.miza.append((karta,iidx))
		
		# print stanje.miza
		
		zmagovalec_idx = kdo_je_zmagal(stanje.miza)
		
		# doublecheck
		zmagovalec_m_ind_idx = idxMaxKarte([k[0] for k in stanje.miza])
		zmagovalec_m_idx = stanje.miza[zmagovalec_m_ind_idx][1]
		if zmagovalec_idx != zmagovalec_m_idx: raise Exception("Nekdo ne zna stet")
		
		stanje.igralci[zmagovalec_idx][0].stat['zmaga'] += 1
		
		koncna_miza = [x[0] for x in stanje.miza]

		if verbose:
			for i in range(4):
				# katero karto ima i-ti igralec na mizi?
				karta = [x[0] for x in stanje.miza if x[1]==i][0]
				print (str(karta).ljust(3)+('*' if i==zmagovalec_m_idx else '')).center(COL_WIDTH),
			print '   (%.1fs)' % (time.time()-timer)
			timer = time.time()
		
		for iidx in [(i+stanje.zacetni_idx) % 4 for i in range(4)]:
			igralec = stanje.igralci[iidx]
			t_b = time.time()
			igralec[0].povzetek(stanje.zacetni_idx, koncna_miza, zmagovalec_idx)
			igralec[0].readexpect('MHM')
			igralec[0].stat['time'] += time.time() - t_b
		
		stanje.igralci[zmagovalec_idx][3] |= set(koncna_miza)
		stanje.zacetni_idx = zmagovalec_idx
	
	# ocenjevanje
	# igralci
	glavni = set([stanje.glavni_idx, stanje.soigralec_idx])
	ostali = set(range(4)) - glavni
	
	# tocke
	tA = sum([steviloTock(stanje.igralci[idx][3]) for idx in glavni])
	tB = sum([steviloTock(stanje.igralci[idx][3]) for idx in ostali]) + steviloTock(stanje.preostanek_talona)
	tt = (tA - tB)//2 + (1 if tA > tB else -1) * stanje.tip_igre.vrednost()
	tocke = [tt if idx in glavni else 0 for idx in range(4)]
	
	# koncaj igro
	for i in range(4):
		stanje.igralci[i][0].konec('konec igre')
		stanje.igralci[i][0].stat['pts'] += tocke[i]
	
	# premecemo tocke list da ustreza accept zaporedju
	rt = [(stanje.igralci[i][5],tocke[i]) for i in range(4)]
	rt.sort()
	return [x[1] for x in rt]

def main(stIger, logIgre=[], xlsLog=[], noncum=False, clientProgs=[]):
	"""
	Zazene streznik, nato pa se vse kliente, ki so navedeni v <clientProgs>
	(z executable filenameom). Pocaka, da se povezejo vsi klienti, nato
	odigra <stIger> iger.
	"""
	print "Cakam na vse igralce..."
	
	random.seed(1)		# !!!
	
	ls = Socket()
	ls.listen(1179)
	
	clientOuts = [open("stdout_client%d.txt" % i, "w") for i in range(1,len(clientProgs)+1)]
	for clientProg, clientOut in zip(clientProgs, clientOuts):
		if clientProg.split()[0].endswith(".py"):
			clientProg = "python.exe " + clientProg
		subprocess.Popen(clientProg, stdout=clientOut, stderr=clientOut)
	
	clients = []
	for i in range(4):
		clients.append(ls.accept())
		print '  ... povezal se je igralec "%s".' % clients[-1][1]
	clients = [(s,str(q)+':'+n,q) for (q,(s,n)) in enumerate(clients)]

	logFile = None
	if not logIgre==[]:
		logFile = open(logIgre[0], 'w')
		logFile.write(','.join([c[1] for c in clients]) + '\n')
	
	xlsFile, xlsWB, xlsWS, xlsLN = None, None, None, 0
	if not xlsLog==[]:
		import xlwt
		xlsFile = xlsLog[0]
		xlsWB = xlwt.Workbook()
		xlsWS = xlsWB.add_sheet('Rezultati')
		for y in range(4):
			xlsWS.write(xlsLN, y, clients[y][1])
		xlsLN += 1
	
	tocke = [0,0,0,0]

	#print "Igralci: ", [x[1] for x in clients]
	
	cperm = itertools.cycle(permutationsX(clients))
	for i in xrange(stIger):
		try:
			premesani_igralci = cperm.next()
			trenutne_tocke = run_game(premesani_igralci)
			tocke = map(lambda x,y:x+y, tocke, trenutne_tocke)
			
			if verbose: print "="*((COL_WIDTH+1)*4)
			print ' '.join(str(t).center(COL_WIDTH) for t in [trenutne_tocke[igr[2]] for igr in premesani_igralci])
			if verbose: print ''
			if logFile:
				d = tocke
				if noncum: d = trenutne_tocke
				logFile.write(','.join([str(t) for t in d]) + '\n')
				logFile.flush()
			if xlsWS:
				d = trenutne_tocke if noncum else tocke
				for y in range(4):
					xlsWS.write(xlsLN, y, d[y])
				xlsLN += 1
		except IOError:
			print 'Igra prekinjena zaradi I/O napake:'
			traceback.print_exc()
			break
	print "*"*((COL_WIDTH+1)*4)
	print ' '.join(c[1][:COL_WIDTH].center(COL_WIDTH) for c in clients)
	for stat, key in (('tocke','pts'),('st. igranj','glavni'),('pobranih stihov','zmaga'), ('cas','time')):
		print ' '.join(str(c[0].stat[key]).center(COL_WIDTH) for c in clients), "   ("+stat+")"
	print ''

	try:	[c[0].adijo() for c in clients]
	except: pass
	
	if logFile: logFile.close()
	if xlsWB:
		xlsWB.save(xlsFile)

	for clientOut in clientOuts:
		clientOut.flush() # naj raje OS dokoncno zapre fajle

CMDLINE_HELP = """
Sintaksa:
  python tarok_eval_cli.py -n <SteviloIger> -v [-c <clientExe>] [-c <clientExe2>] ... [-c <clientExeN>]

Parametri:
  -n <SteviloIger> -- Stevilo iger, ki naj jih odigrajo igralci
  -v -- Verbose. Izpisuj potek posameznih iger.
  -c <clientExe> -- Argument te oblike se lahko pojavi veckrat. Za vsako
      od pojavitev streznik na lokalnem racunalniku zazene navedenega clienta
      in ga poveze nase.
  -l <logFile> -- Logira rezultate vseh iger v podani CSV file.
  -x <logFile> -- Logira rezultate vseh iger v podani XLS (Excel) file.
  -p -- Logira rezultate posameznih iger, ne kumulativnih tock.
"""

if __name__ == "__main__":
	print "Tarok testni streznik, razlicica $Rev: 4756 $"
	try:
		opts, unparsed = getopt.getopt(sys.argv[1:], "n:vc:l:x:p")
		assert unparsed==[], "Unknown option(s): "+', '.join(unparsed)
		verbose = any(o[0]=='-v' for o in opts)
		clients = [o[1] for o in opts if o[0]=='-c']
		stIger = int([o[1] for o in opts if o[0]=='-n'][0])
		logIgre = [o[1] for o in opts if o[0]=='-l'][:1]
		xlsLog = [o[1] for o in opts if o[0]=='-x'][:1]
		noncum = any(o[0]=='-p' for o in opts)
	except:	
		print CMDLINE_HELP
		#traceback.print_exc()
		sys.exit(1)

	main(stIger, logIgre, xlsLog, noncum, clients)
